///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2019 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

package com.crankuptheamps.client;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.math.BigInteger;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.StandardCharsets;
import java.nio.file.FileSystems;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.lang.Long;
import java.util.*;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import com.crankuptheamps.client.fields.Field;
import com.crankuptheamps.client.fields.BookmarkField;
import com.crankuptheamps.client.exception.*;
/** LoggedBookmarkStore implements a sequentially written log
 * of incoming and discarded messages. This store tracks every
 * bookmark processed in a file. An application should periodically call
 * {@link #prune} to manage the size of the file by removing
 * outdated entries.
 */
public class LoggedBookmarkStore implements BookmarkStore
{
    // Each entry begins with a single byte indicating the type of entry:
    // a new bookmark, or a discard of a previous one.
    static final byte ENTRY_BOOKMARK  = (byte)'b';
    static final byte ENTRY_DISCARD   = (byte)'d';
    static final byte ENTRY_PERSISTED = (byte)'p';

    // We use an in-memory array to remember the order in which bookmarks
    // arrive on a given subscription.  When the user has discard()ed all of the
    // bookmarks up to a certain bookmark B, we log B, and return it the next time
    // someone calls getMostRecent().
    static class Subscription implements com.crankuptheamps.client.Subscription
    {
        // This subscription's ID.
        Field _sub;

        // The last persisted bookmark if server version is > MIN_MULTI_BOOKMARK_VERSION
        private BookmarkField _lastPersisted;

        /**
         * The last-modified timestamp of the backing bookmark log file, just
         * before recovery is initiated. When this is not null we intend to
         * include it in the list of bookmarks returned by getMostRecentList()
         * until a message is discarded on the sub or the sub state is purged.
         *
         * Originally we thought the recovery timestamp could be stored
         * as the last-persisted for each recovered sub, but it must be
         * stored separately here because setting it as a sub's
         * last-persisted could cause us to loose a persisted ack that
         * is earlier than the backing file's last-modified time.
         */
        protected volatile String _recoveryTimestamp = null;

        static final Field EPOCH_FIELD = new Field(Client.Bookmarks.EPOCH);

        static final CharsetEncoder _encoder = StandardCharsets.UTF_8.newEncoder();
        static final CharsetDecoder _decoder = StandardCharsets.UTF_8.newDecoder();

        // A set of all of the undiscarded entries recovered that
        // had been received after the one returned by getMostRecent().
        // When these come in again, we won't log them a 2nd time.
        HashMap<Field, Long> _recovered = new HashMap<Field, Long>();

        // A map of active bookmark indices to their new location after a reconnect
        HashMap<Long, Long> _relocatedActiveBookmarks = new HashMap<Long, Long>();

        BookmarkRingBuffer _ring = new BookmarkRingBuffer();

        // The per-subscription memory of what we've seen from publishers
        HashMap<Long, Long> _publishers = new HashMap<Long, Long>();

        // The store that we log to when the time comes.
        LoggedBookmarkStore _parent;

        // The Subscription lock
        private final Lock       _lock           = new ReentrantLock();

        public Subscription()
        {
            // Default starting point for last persisted is EPOCH
            _lastPersisted = new BookmarkField();
            _lastPersisted.copyFrom(EPOCH_FIELD);
        }

        /**
         * Reset the state of this subscription object such that it can be
         * returned to the pool for reuse.
         */
        public void reset() {
            _sub.reset();
            _lastPersisted.copyFrom(EPOCH_FIELD);
            _recoveryTimestamp = null;
            _recovered.clear();
            _relocatedActiveBookmarks.clear();
            _ring.reset();
            _publishers.clear();
        }

        public void init(Field subscriptionId, LoggedBookmarkStore parent) throws IOException
        {
            _sub = subscriptionId.copy();
            _ring.setSubId(_sub);
            _parent = parent;
        }

        public String getRecoveryTimestamp()
        {
            return _recoveryTimestamp;
        }

        protected void setRecoveryTimestamp(String rts)
        {
            _recoveryTimestamp = rts;
        }

        public long log(BookmarkField bookmark) throws IOException
        {
            _lock.lock();
            try {
                // Check to see if this is a recovered bookmark.
                // If so, re-log in the new order.
                final Long recoveredValue = _recovered.remove(bookmark);
                if (recoveredValue != null)
                {
                    long seq = _ring.relog(recoveredValue, bookmark);
                    while (seq == 0) {
                        _lock.unlock();
                        try {
                            _ring.checkResize();
                        }
                        finally {
                            _lock.lock();
                        }
                        seq = _ring.relog(recoveredValue, bookmark);
                    }
                    _relocatedActiveBookmarks.put(recoveredValue, seq);
                    return seq;
                }

                // Add this entry onto our list to remember in order.
                long seq = _ring.log(bookmark);
                while (seq == 0) {
                    _lock.unlock();
                    try {
                        _ring.checkResize();
                    }
                    finally {
                        _lock.lock();
                    }
                    seq = _ring.log(bookmark);
                }
                return seq;
            }
            finally {
                _lock.unlock();
            }
        }

        public void discard(long index) throws IOException
        {
            _lock.lock();
            try {
                long idx = index;
                Long newIndex = _relocatedActiveBookmarks.remove(index);
                // In case something's been recovered more than once, get
                // full indirection
                while (newIndex != null)
                {
                    _ring.discard(idx);
                    idx = newIndex;
                    newIndex = _relocatedActiveBookmarks.remove(newIndex);
                }
                BookmarkRingBuffer.Entry entry = _ring.getByIndex(idx);

                if(entry == null || !entry.isActive())
                {
                    return;
                }

                _parent.write(_sub, LoggedBookmarkStore.ENTRY_DISCARD,
                      entry.getBookmark());
                if (_ring.discard(idx)) {
                    _parent._recentChanged = true;
                    _recoveryTimestamp = null;
                }
            }
            finally {
                _lock.unlock();
            }
        }

        /** Check to see if this message is older than the most recent one seen,
         * and if it is, then check to see if it is discarded.
         */
        public boolean isDiscarded(BookmarkField bookmark) throws IOException
        {
            _lock.lock();
            try {
                Long recoveredIndex = _recovered.get(bookmark);
                if (recoveredIndex != null)
                {
                    BookmarkRingBuffer.Entry entry = _ring.getByIndex(recoveredIndex);
                    // Need current active state
                    boolean active = (entry == null) ? false : entry.isActive();
                    // It's there, not yet discarded
                    if (active)
                    {
                        return false;
                    }
                    // It's either not there because it was discarded and persisted
                    // and we cleaned it up on a previous incoming bookmark or it's
                    // still there and discarded and persisted AND we don't have
                    // anything new since recovery started.
                    if ((entry == null || entry.isPersisted() ) &&
                        _ring.getStartIndex() == BookmarkRingBuffer.UNSET_INDEX)
                    {
                        _ring.discard(recoveredIndex); // Updates most recent
                        _parent.write(_sub, LoggedBookmarkStore.ENTRY_DISCARD, bookmark);
                    }
                    else // It's not persisted so relog
                    {
                        // We don't need to save the new index from relog since that's
                        // only used in discard and this message is already discarded
                        long seq = _ring.relog(recoveredIndex, bookmark);
                        while (seq == 0) {
                            _lock.unlock();
                            try {
                                _ring.checkResize();
                            }
                            finally {
                                _lock.lock();
                            }
                            seq = _ring.relog(recoveredIndex, bookmark);
                        }
                        _parent.write(_sub, LoggedBookmarkStore.ENTRY_BOOKMARK, bookmark);
                        _parent.write(_sub, LoggedBookmarkStore.ENTRY_DISCARD, bookmark);
                    }
                    _recovered.remove(bookmark);
                    return true;
                }
                long publisher = bookmark.getPublisherId();
                long sequence = bookmark.getSequenceNumber();
                if(!_publishers.containsKey(publisher) ||
                    unsignedLongLess(_publishers.get(publisher), sequence))
                {
                    _publishers.put(publisher, sequence);
                    return false;
                }
                
                // If we are in the middle of performing a recovery, then
                // the return value of this method isn't used anymore (it's
                // just getting called to setup the _publishers map),
                // so don't waste time in _ring.find() below.
                if (_parent._recovering) {
                    return false;
                }
                
                // During fail-over scenarios, we'll see out of order
                // bookmarks arrive, either (a) because we're replaying or (b)
                // because a publisher has cut over, and we've cut over to a new server.
                // Scan the list to see if we have a match.
                BookmarkRingBuffer.Entry entry = _ring.find(bookmark);

                if(entry != null)
                {
                    return !entry.isActive();
                }

                return true; // message is totally discarded
            }
            finally {
                _lock.unlock();
            }
        }

        public Field getMostRecent()
        {
            return getMostRecent(true);
        }

        public Field getMostRecent(boolean updateRecovery_)
        {
            _lock.lock();
            try {
                // when this is called, we'll take a moment to update the list of things
                // recovered, so we don't accidentally log anything we ought not to.
                if (updateRecovery_ && _parent._recentChanged) updateRecovery();
                return _ring.getLastDiscarded();
            }
            finally {
                _lock.unlock();
            }
        }

        /** publisherId_ and sequence are signed, but the actual id's are
         * unsigned, so we have to do some work to create a valid string.
         */
        private static String convertUnsignedLongToString(long publisherId_)
        {
            final BigInteger offset = BigInteger.valueOf(Long.MAX_VALUE)
                    .shiftLeft(1).add(BigInteger.valueOf(2));
            if(publisherId_<0)
            {
                return offset.add(BigInteger.valueOf(publisherId_)).toString();
            }
            else
            {
                return Long.toString(publisherId_);
            }
        }

        /** publisherId_ and sequence are signed, but the actual id's are
         * unsigned, so we have to do some work to compare them.
         */
        private static boolean unsignedLongLess(long seqLeft_, long seqRight_)
        {
            final BigInteger offset = BigInteger.valueOf(Long.MAX_VALUE)
                    .shiftLeft(1).add(BigInteger.valueOf(2));
            if(seqLeft_<0 || seqRight_<0)
            {
                BigInteger left = offset.add(BigInteger.valueOf(seqLeft_));
                BigInteger right = offset.add(BigInteger.valueOf(seqRight_));
                return left.compareTo(right) < 0;
            }
            else
            {
                return seqLeft_ < seqRight_;
            }
        }

        private static boolean unsignedLongLessEqual(long seqLeft_, long seqRight_)
        {
            final BigInteger offset = BigInteger.valueOf(Long.MAX_VALUE)
                    .shiftLeft(1).add(BigInteger.valueOf(2));
            if(seqLeft_<0 || seqRight_<0)
            {
                BigInteger left = offset.add(BigInteger.valueOf(seqLeft_));
                BigInteger right = offset.add(BigInteger.valueOf(seqRight_));
                return left.compareTo(right) <= 0;
            }
            else
            {
                return seqLeft_ <= seqRight_;
            }
        }

        public Field getMostRecentList()
        {
            _lock.lock();
            try {
                // when this is called, we'll take a moment to update the list of things
                // recovered, so we don't accidentally log anything we ought not to.
                updateRecovery();
                BookmarkField lastDiscarded = (BookmarkField)_ring.getLastDiscarded();
                boolean useLastDiscarded = (lastDiscarded != null &&
                                            !lastDiscarded.isNull());
                long lastDiscardedPub = 0;
                long lastDiscardedSeq = 0;
                boolean useLastPersisted = (_lastPersisted != null &&
                                            !_lastPersisted.isNull());
                long lastPersistedPub = 0;
                long lastPersistedSeq = 0;
                if (useLastPersisted)
                {
                    lastPersistedPub = _lastPersisted.getPublisherId();
                    lastPersistedSeq = _lastPersisted.getSequenceNumber();
                }
                if (useLastDiscarded)
                {
                    if (_ring.isEmpty() && useLastPersisted)
                    {
                        useLastDiscarded = false;
                    }
                    else
                    {
                        lastDiscardedPub = lastDiscarded.getPublisherId();
                        lastDiscardedSeq = lastDiscarded.getSequenceNumber();
                        // Only use one if they are same publisher
                        if (useLastPersisted && (lastDiscardedPub == lastPersistedPub))
                        {
                            useLastDiscarded = (lastDiscardedSeq < lastPersistedSeq);
                            useLastPersisted = !useLastDiscarded;
                        }
                    }
                }
                StringBuilder recentStr = new StringBuilder();
                BookmarkField recentList = new BookmarkField();
                if (_recoveryTimestamp != null)
                {
                    recentStr.append(_recoveryTimestamp);
                    recentStr.append(',');
                }
                if (useLastDiscarded)
                {
                    recentStr.append((lastDiscarded).getValue(_decoder));
                }
                // If we don't have a last persisted or a last discarded OR we are
                // expecting persisted acks but haven't received one yet, then we
                // should try to build a list of bookmarks based on publishers we
                // have seen so far, if any, or return EPOCH.
                if ((!useLastPersisted && !useLastDiscarded) ||
                     (_lastPersisted != null &&
                      _lastPersisted.equals(EPOCH_FIELD)))
                {
                    if (_publishers.isEmpty())
                    {
                        // Set last persisted to EPOCH and return it
                        if (_lastPersisted == null) {
                            _lastPersisted = new BookmarkField();
                            _lastPersisted.copyFrom(EPOCH_FIELD);
                        }
                        return _lastPersisted;
                    }
                    if (recentStr.length() > 0) recentStr.setLength(0);
                    Iterator it = _publishers.entrySet().iterator();
                    while (it.hasNext())
                    {
                        if (recentStr.length() > 0) recentStr.append(',');
                        Map.Entry pairs = (Map.Entry)it.next();
                        long pubId = (Long)pairs.getKey();
                        if (pubId == 0) continue;
                        long seq = (Long)pairs.getValue();
                        recentStr.append(convertUnsignedLongToString(pubId)).append(BookmarkField.SEPARATOR_CHAR);
                        recentStr.append(convertUnsignedLongToString(seq)).append(BookmarkField.SEPARATOR_CHAR);
                    }
                    recentList.setValue(recentStr.toString(), _encoder);
                    return recentList;
                }
                if (useLastPersisted)
                {
                    if (recentStr.length() > 0) recentStr.append(',');
                    recentStr.append(_lastPersisted.getValue(_decoder));
                }
                recentList.setValue(recentStr.toString(), _encoder);
                return recentList;
            }
            finally {
                _lock.unlock();
            }
        }

        private void updateRecovery()
        {
            _recovered.clear();
            long end = _ring.getEndIndex();
            for(long index = _ring.setRecovery(); index< end; ++index)
            {
                BookmarkRingBuffer.Entry entry = _ring.getByIndex(index);
                if(entry != null && entry._bookmark != null && !entry._bookmark.isNull())
                {
                    _recovered.put(entry.getBookmark(), index);
                }
            }
        }

        private ArrayList<BookmarkRingBuffer.Entry> getRecoveryEntries()
        {
            return _ring.getRecoveryEntries();
        }

        private void getActiveEntries(ArrayList<BookmarkRingBuffer.Entry> entryList_)
        {
            _ring.getRecoveryEntries(entryList_);
        }

        public void persisted(long bookmark) throws IOException
        {
            _lock.lock();
            try {
                BookmarkRingBuffer.Entry entry = _ring.getByIndex(bookmark);
                if (entry == null)
                    return;
                BookmarkField bookmarkField = entry.getBookmark();
                _parent.write(_sub, ENTRY_PERSISTED, bookmarkField);
                _ring.persisted(bookmark);
            }
            finally {
                _lock.unlock();
            }
        }

        public void persisted(BookmarkField bookmark) throws IOException
        {
            _lock.lock();
            try {
                if (bookmark == null || bookmark.isNull()) return;
                _parent.write(_sub, ENTRY_PERSISTED, bookmark);
                _ring.persisted(bookmark);
            }
            finally {
                _lock.unlock();
            }
        }

        public void setLastPersisted(long bookmark) throws IOException
        {
            _lock.lock();
            try {
                BookmarkRingBuffer.Entry entry = _ring.getByIndex(bookmark);
                if (entry == null)
                    return;
                BookmarkField bookmarkField = entry.getBookmark();
                _parent.write(_sub, ENTRY_PERSISTED, bookmarkField);
                if (_lastPersisted != null) _lastPersisted.reset();
                _lastPersisted = bookmarkField.copy();
            }
            finally {
                _lock.unlock();
            }
        }

        public void setLastPersisted(BookmarkField bookmark) throws IOException
        {
            _lock.lock();
            try {
                if (bookmark == null || bookmark.isNull()) return;

                if (bookmark.equals(_lastPersisted)) return;

                if (bookmark.isTimestamp())
                {
                    _parent._recentChanged = true;;
                    if (_lastPersisted != null) _lastPersisted.reset();
                    _lastPersisted = bookmark.copy();
                    _parent.write(_sub, ENTRY_PERSISTED, bookmark);
                    return;
                }

                long publisherId = bookmark.getPublisherId();
                if (_lastPersisted != null && !_ring.persistedAcks() && 
                    publisherId == _lastPersisted.getPublisherId() &&
                    bookmark.getSequenceNumber() <= _lastPersisted.getSequenceNumber())
                {
                    return;
                }
                if (publisherId == _ring.getLastDiscarded().getPublisherId()
                    || _ring.isEmpty())
                    _parent._recentChanged = true;;
                if (_lastPersisted != null) _lastPersisted.reset();
                _lastPersisted = bookmark.copy();
                _parent.write(_sub, ENTRY_PERSISTED, bookmark);
            }
            finally {
                _lock.unlock();
            }
        }

        public void noPersistedAcks()
        {
            _lock.lock();
            try {
                if (_parent.getServerVersion() < Client.MIN_PERSISTED_BOOKMARK_VERSION ||
                    _parent._noPersistedAcks.contains(_sub))
                    _lastPersisted = null;
                _ring.noPersistedAcks();
            }
            finally {
                _lock.unlock();
            }
        }

        public void setPersistedAcks()
        {
            _lock.lock();
            try {
                _ring.setPersistedAcks();
                if (_lastPersisted == null)
                {
                    _lastPersisted = new BookmarkField();
                }
                // Default starting point for last persisted is EPOCH
                _lastPersisted.copyFrom(EPOCH_FIELD);
            }
            finally {
                _lock.unlock();
            }
        }

        public long getOldestBookmarkSeq()
        {
            _lock.lock();
            try {
                return _ring.getStartIndex();
            }
            finally {
                _lock.unlock();
            }
        }

        /**
         * Call on a Subscription object just after recovery is performed to
         * convert logged entries into recovery entries and set the publishers
         * cache state to the earliest sequence seen for each publisher minus
         * one.
         *
         * NOTE: If after recovery lastDiscarded for this sub is null (i.e.
         * nothing was discarded) and lastPersisted is null or EPOCH (i.e.
         * no persisted ack was recorded in the log), then we can throw away
         * this subscription state and return this Subscription object to the
         * pool.
         *
         * @return Indicates whether this subscription's state was reset
         *      (because it's undeeded due to the lack of any discards and
         *      persisted acks) and this object should be returned to the
         *      Subscription pool.
         */
        public boolean justRecovered()
        {
            _lock.lock();
            try {
                BookmarkField ld = _ring.getLastDiscarded();
                if ((ld == null || ld.isNull() || ld.equals(EPOCH_FIELD))
                        && (_lastPersisted == null || _lastPersisted.isNull()
                                || _lastPersisted.equals(EPOCH_FIELD)))
                {
                    // Reset this sub for reuse.
                    reset();
                    return true;
                }

                updateRecovery();
                setPublishersToDiscarded(getRecoveryEntries(), _publishers);
            }
            finally {
                _lock.unlock();
            }

            return false;
        }

        public static void setPublishersToDiscarded(
                                    List<BookmarkRingBuffer.Entry> active,
                                    Map<Long, Long> publishers)
        {
            if (active == null || publishers == null || publishers.isEmpty()) return;
            Iterator<BookmarkRingBuffer.Entry> it = active.iterator();
            while (it.hasNext()) {
                BookmarkField bf = it.next().getBookmark();
                long seq = bf.getSequenceNumber();
                if (seq == 0) continue;
                long publisher = bf.getPublisherId();
                Long pubSeq = publishers.get(publisher);
                if (pubSeq != null && unsignedLongLessEqual(seq, pubSeq))
                    publishers.put(publisher, seq - 1);
            }
        }

        public void setResizeHandler(BookmarkStoreResizeHandler handler, BookmarkStore store)
        {
            _ring.setResizeHandler(handler, store);
        }

        /**
         * Lock self's internal lock.
         * Used by LoggedBookmarkStore when gathering Subscriptions to prune.
         */
        void lock()
        {
            _lock.lock();
        }

        /**
         * Unlock self's internal lock.
         * Used by LoggedBookmarkStore when finishing a prune() operation.
         */
        void unlock()
        {
            _lock.unlock();
        }
    }

    HashMap<Field, Subscription> _subs = new HashMap<Field, Subscription>();
    HashSet<Field> _noPersistedAcks = new HashSet<Field>();
    RandomAccessFile _file;
    String _fileName;
    volatile boolean _recovering = false;
    volatile boolean _recentChanged = true;
    Pool<Subscription> _pool;
    BookmarkStoreResizeHandler _resizeHandler = null;
    private int _serverVersion = Client.MIN_MULTI_BOOKMARK_VERSION;
    static final CharsetDecoder _decoder = StandardCharsets.UTF_8.newDecoder();
    // The Store lock
    private final Lock       _lock           = new ReentrantLock();
    private final Lock       _subsLock           = new ReentrantLock();

    /**
     * Same as calling LoggedBookmarkStore(path, 1, false).
     * 
     * @see #LoggedBookmarkStore(String, int, boolean)
     */
    public LoggedBookmarkStore(String path) throws IOException
    {
        this(path, 1);
    }

    /**
     * Same as calling LoggedBookmarkStore(path, targetNumberOfSubscriptions, false).
     * 
     * @see #LoggedBookmarkStore(String, int, boolean)
     */
    public LoggedBookmarkStore(String path, int targetNumberOfSubscriptions) throws IOException
    {
        this(path, targetNumberOfSubscriptions, false);
    }

    /**
     * A file-backed bookmark store implementation that fully supports
     * discarding messages in an order different from the order they arrived
     * (i.e. out-of-order discards) and fail-over to a replicated server.
     * All messages must eventually be discarded, otherwise memory usage will
     * become proportional to the number of messages received on a
     * bookmark subscription since the first undiscarded message.
     * This implementation requires that the prune() method be called
     * periodically to removed unneeded discarded bookmarks from the
     * backing-file, otherwise it will continue to grow without bound.
     * The prune() method is thread-safe and can be called from any thread.
     *
     * @param path The path to the backing bookmark log file. 
     * @param targetNumberOfSubscriptions The initial capacity for the number
     *        of bookmark subscriptions you anticipate creating on the AMPS
     *        client instance that this bookmark store is registered on. This
     *        will grow as needed if more subscriptions are created than
     *        anticipated.
     * @param useLastModifiedTime Indicates whether the recovery timestamp
     *        feature should be used. If true, the last-modified time of the
     *        backing file is included (as an AMPS timestamp bookmark in a
     *        comma-separated list of bookmarks) when getMostRecent() is called
     *        after recovering from a bookmark file. This feature could be
     *        useful if you have an infrequently run process that is run on a
     *        schedule that is longer than the AMPS server keeps messages in
     *        its transaction log. When this process is started and recovers
     *        a bookmark log full of old bookmarks that are no longer available
     *        using the MOST_RECENT bookmark indicator, the recovery timestamp
     *        will cause the bookmark subscription to begin at the start of the
     *        transaction log (e.g. EPOCH), rather than its tail (e.g. NOW).
     *
     * @throws IOException If there is a problem creating, reading, or writing
     *         the backing file.
     */
    public LoggedBookmarkStore(String path,
                               int targetNumberOfSubscriptions,
                               boolean useLastModifiedTime) throws IOException
    {
        File bmLogFile = new File(path);
        String recoveryTimestamp = null;
        if (useLastModifiedTime && bmLogFile.exists()) {
            Date lastMod = new Date(bmLogFile.lastModified());
            TimeZone tz = TimeZone.getTimeZone("UTC");
            DateFormat df = new SimpleDateFormat("yyyyMMdd'T'HHmmss'Z'");
            df.setTimeZone(tz);
            recoveryTimestamp = df.format(lastMod);
        }

        _pool = new Pool<Subscription>(Subscription.class, targetNumberOfSubscriptions);
        _file = new RandomAccessFile(path, "rw");
        _fileName = path;

        try
        {
            recover();

            if (recoveryTimestamp != null)
            {
                // Initialize the recovery timestamp on each recovered sub.
                for (Subscription sub: _subs.values())
                {
                    sub.setRecoveryTimestamp(recoveryTimestamp);
                }
            }
        }
        catch(IOException ioex)
        {
            try
            {
                _file.close();
            }
            catch (IOException ignoreTheCloseException)
            {
            }
            finally
            {
                _file = null;
            }
            throw ioex;
        }
    }

    /**
     *  Remove outdated entries in the bookmark store.
     *  This function creates a temporary file, copies current entries to
     *  that file, and then replaces the current file with the temporary
     *  file.
     *  @throws IOException Thrown when an operation on the file fails.
     *  @throws StoreException Thrown when any other operation fails with details on the failure.
     */
    public void prune() throws IOException, StoreException
    {
        String name = _fileName + ".tmp";
        prune(name);
    }

    /**
     *  Remove outdated entries in the bookmark store.
     *  This function creates a temporary file, copies current entries to
     *  that file, and then replaces the current file with the temporary
     *  file.
     *  @param tmpFileName_  The name of the temporary file.
     *  @throws IOException Thrown when an operation on the file fails.
     *  @throws StoreException Thrown when any other operation fails with details on the failure.
     */

    public void prune(String tmpFileName_) throws IOException, StoreException
    {
        List<Map.Entry<Field,Subscription>> subs = null;
        _subsLock.lock();
        try
        {
            try
            {
                subs = _lockSubs();
                prune(tmpFileName_,subs);
            }
            finally
            {
                if (subs != null) _unlockSubs(subs);
            }
        }
        finally
        {
            _subsLock.unlock();
        }
    }

    /**
     *  Remove outdated entries in the bookmark store.
     *  This function is used internally and processes a list of subscriptions
     *  whose locks are already held, so that the locking order of subscription
     *  then store can be preserved.
     *  @param tmpFileName_ The name of the temporary file.
     *  @param subs_        The list of Subscriptions to prune.
     */
    private void prune(String tmpFileName_,
                       List<Map.Entry<Field,Subscription>> subs_)
      throws IOException, StoreException
    {
        RandomAccessFile file = new RandomAccessFile(tmpFileName_, "rw");
        _lock.lock();
        try
        {
            if(_file == null)
            {
                throw new StoreException("Store not open.");
            }
            if (!_recentChanged) return;
            try
            {
                file.getChannel().force(false);
                StringBuilder bookmarkBuilder = new StringBuilder(64);
                Field bookmark = new Field();
                for (Map.Entry<Field,Subscription> sub : subs_)
                {
                    Field subId = sub.getKey();
                    assert(subId != null);
                    Subscription subscription = sub.getValue();
                    BookmarkField recent = (BookmarkField)subscription.getMostRecent(false);
                    if (recent.isNull()) {
                        recent.copyFrom(subscription.EPOCH_FIELD);
                    }
                    long recentPub = recent.getPublisherId();
                    ArrayList<BookmarkRingBuffer.Entry> recovered = subscription.getRecoveryEntries();
                    HashMap<Long, Long> publishers = new HashMap<Long, Long>(subscription._publishers);
                    Subscription.setPublishersToDiscarded(recovered, publishers);

                    // First write the highest discard for each publisher other
                    // than most recent.
                    Iterator<Map.Entry<Long, Long>> pubIter = publishers.entrySet().iterator();
                    while (pubIter.hasNext()) {
                        Map.Entry<Long, Long> publisher = pubIter.next();
                        long pubId = publisher.getKey();
                        long pubSeq = publisher.getValue();
                        if (pubId == 0 || pubSeq == 0 || pubId == recentPub)
                            continue;
                        bookmarkBuilder.setLength(0);
                        bookmarkBuilder.append(Subscription.convertUnsignedLongToString(pubId))
                                       .append(BookmarkField.SEPARATOR_CHAR)
                                       .append(Subscription.convertUnsignedLongToString(pubSeq))
                                       .append(BookmarkField.SEPARATOR_CHAR);
                        bookmark.set(bookmarkBuilder.toString().getBytes(StandardCharsets.UTF_8), 0, bookmarkBuilder.length());
                        writeBookmarkToFile(file, subId, bookmark, ENTRY_BOOKMARK);
                        writeBookmarkToFile(file, subId, bookmark, ENTRY_DISCARD);
                    }
                    // Now write the most recent
                    if (recent.length > 1) {
                        writeBookmarkToFile(file, subId, recent, ENTRY_BOOKMARK);
                        writeBookmarkToFile(file, subId, recent, ENTRY_DISCARD);
                    }
                    // Now write the last persisted
                    if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION)
                    {
                        if (subscription._lastPersisted != null
                            && subscription._lastPersisted.length > 1)
                        {
                            writeBookmarkToFile(file, subId,
                                                subscription._lastPersisted,
                                                ENTRY_PERSISTED);
                        }
                    }
                    // Finally, write what's left in entries.
                    if (recovered == null)
                    {
                        recovered = new ArrayList<BookmarkRingBuffer.Entry>();
                    }
                    subscription.getActiveEntries(recovered);
                    for (BookmarkRingBuffer.Entry entry : recovered)
                    {
                        BookmarkField entryBookmark = entry.getBookmark();
                        if (entryBookmark == null
                            || entryBookmark.isNull())
                        {
                            continue;
                        }
                        writeBookmarkToFile(file, subId, entryBookmark,
                                            ENTRY_BOOKMARK);
                        if (!entry.isActive())
                            writeBookmarkToFile(file, subId,
                                                entryBookmark,
                                                ENTRY_DISCARD);
                        // Persisted on entry only matters for specific server versions
                        if (_serverVersion <
                                           Client.MIN_MULTI_BOOKMARK_VERSION &&
                            _serverVersion >=
                                       Client.MIN_PERSISTED_BOOKMARK_VERSION &&
                            entry.isPersisted())
                        {
                            writeBookmarkToFile(file, subId,
                                                entryBookmark,
                                                ENTRY_PERSISTED);
                        }
                    }
                }
            }
            // If we end up in catch because of an IOException, need to clean tmp file
            catch (IOException e)
            {
                file.close();
                File tmp = new File(tmpFileName_);
                // We can ignore a failure of delete here, prune has failed
                tmp.delete();
                throw new StoreException("Failed attempting to prune file " + _fileName + " to " + tmpFileName_, e);
            }
            // Close the files, delete the original, move the pruned file
            file.close();
            _file.close();
            _file = null;
            int retries = 0;
            File origTmp = new File(_fileName);
            while (retries++ < 3)
            {
                if (!origTmp.delete())
                {
                    if (retries >= 3) throw new StoreException(
                            "Failed to delete original file " + _fileName
                            + " after completing prune to " + tmpFileName_);
                }
                else 
                {
                    break;
                }
            }
            // Let file system catch up
            while (origTmp.exists())
            {
                try {
                    Thread.sleep(100L);
                } catch (InterruptedException e) {
                    // Ignore
                }
            }
            retries = 0;
            while (retries++ < 3)
            {
                File tmp = new File(tmpFileName_);
                if (!tmp.renameTo(origTmp))
                {
                    if (retries >= 3) throw new StoreException(
                            "Failed to rename pruned file " + tmpFileName_
                            + " to original file name: " + _fileName);
                    try { Thread.sleep(50); }
                    catch (InterruptedException ex) { }
                }
                else 
                {
                    break;
                }
            }
            _file = new RandomAccessFile(_fileName, "rw");
            if (_file.length() > 0) _file.seek(_file.length());
            _recentChanged = false;
        }
        finally {
            _lock.unlock();
        }
    }

    private void writeBookmarkToFile(RandomAccessFile file_, Field subId_, Field bookmark_, byte entry_) throws IOException
    {
        file_.writeByte(subId_.length);
        file_.write(subId_.buffer, subId_.position, subId_.length);
        file_.writeByte(entry_);
        file_.writeByte(bookmark_.length);
        file_.write(bookmark_.buffer, bookmark_.position, bookmark_.length);
        file_.writeByte((byte)'\n');
    }

    void write(Field sub, byte entry, Field data) throws IOException
    {
        _lock.lock();
        try {
            if(!_recovering)
            {
                writeBookmarkToFile(_file, sub, data, entry);
            }
        }
        finally {
            _lock.unlock();
        }
    }

    void write(Field sub, byte entry, long data) throws IOException
    {
        _lock.lock();
        try {
            if(!_recovering)
            {
                _file.writeByte(sub.length);
                _file.write(sub.buffer, sub.position, sub.length);
                _file.writeByte(entry);
                _file.writeLong(data);
                _file.writeByte((byte)'\n');
            }
        }
        finally {
            _lock.unlock();
        }
    }

    /**
     * This method is used internally by this {@link BookmarkStore}
     * implementation to recover its state from the bookmark store file.
     * This allows it to know on restart what messages have been discarded by
     * a subscriber and what the most recent message a subscriber should receive
     * from a publisher for a bookmark subscription.
     * 
     * IMPORTANT NOTE: When making changes to this method, please run the
     * manual test case in
     * gfs/sixty/amps-client-java/wip/manual_tests/LargeUnprunedBookmarkStore
     * to ensure proper function in certain hard to automate cases such
     * as ensuring reasonable memory footprint.
     * 
     * @throws IOException If an unrecoverable problem is detected while
     *         processing the bookmark log file.
     */
    private void recover() throws IOException
    {
        _recovering = true;
        HashMap<Field, HashMap<String,Long>> ids = new HashMap<Field, HashMap<String,Long>>();
        HashMap<Field, LinkedList<String>> removeMap = new HashMap<Field, LinkedList<String>>();
        byte[] sub = new byte[255];
        byte[] bookmark = new byte[255];
        long lastGoodPosition = _file.getFilePointer();
        Field subId = new Field();
        BookmarkField bookmarkField = new BookmarkField();
        Long zero = Long.valueOf(0);
        // Create temporary buffered stream for bookmark file. We use a new file
        // channel so that closing the stream later won't close _file.
        Path path = FileSystems.getDefault().getPath(_fileName);
        FileChannel channel = FileChannel.open(path, StandardOpenOption.READ);
        channel.position(lastGoodPosition);
        DataInputStream in = null;
        try
        {
            in = new DataInputStream(
                    new BufferedInputStream(Channels.newInputStream(channel),
                                            8192));
            long position = lastGoodPosition;
            while(lastGoodPosition < _file.length())
            {
                long entryStartPos = position;
                Field subIdCopy = null;
                int subLen = in.readUnsignedByte();
                position++;
                in.readFully(sub, 0, subLen);
                subId.set(sub, 0, subLen);
                position += subLen;
                HashMap<String, Long> subscriptionMap = ids.get(subId);
                if(subscriptionMap == null)
                {
                    subscriptionMap = new HashMap<String, Long>();
                    subIdCopy = subId.copy();
                    ids.put(subIdCopy, subscriptionMap);
                }
                LinkedList<String> toRemove = removeMap.get(subId);
                if(_serverVersion < Client.MIN_MULTI_BOOKMARK_VERSION &&
                        _serverVersion >= Client.MIN_PERSISTED_BOOKMARK_VERSION &&
                        toRemove == null)
                {
                    toRemove = new LinkedList<String>();
                    if (subIdCopy == null) subIdCopy = subId.copy();
                    removeMap.put(subIdCopy, toRemove);
                }
                Subscription subscription = find(subId);
                
                int entryType = in.readUnsignedByte();
                position++;
                switch(entryType)
                {
                case ENTRY_BOOKMARK:
                    int len = in.readUnsignedByte();
                    position++;
                    if(len == -1 || len > (_file.length() - position))
                    {
                        // truncated entry
                        _file.seek(entryStartPos);
                        in.close();
                        return;
                    }
                    in.readFully(bookmark, 0, len);
                    position += len;
                    bookmarkField.set(bookmark, 0, len);
                    String bmStr = bookmarkField.getValue(_decoder);
                    if(subscriptionMap.get(bmStr) != null)
                    {
                        subscription.getMostRecent();
                        subscriptionMap.clear();
                    }
                    if (!subscription.isDiscarded(bookmarkField))
                    {
                        long addedIndex = subscription.log(bookmarkField);
                        subscriptionMap.put(bmStr, addedIndex);
                    }
                    else
                    {
                        subscriptionMap.put(bmStr, zero);
                    }
                    break;

                case ENTRY_DISCARD:
                    len = in.readUnsignedByte();
                    position++;
                    if(len == -1 || len > (_file.length() - position))
                    {
                        // truncated entry
                        _file.seek(entryStartPos);
                        in.close();
                        return;
                    }
                    in.readFully(bookmark, 0, len);
                    position += len;
                    bookmarkField.set(bookmark, 0, len);
                    bmStr = bookmarkField.getValue(_decoder);

                    Long subscriptionMapEntry = subscriptionMap.get(bmStr);
                    if(subscriptionMapEntry != null)
                    {
                        if (_serverVersion < Client.MIN_MULTI_BOOKMARK_VERSION &&
                                _serverVersion >= Client.MIN_PERSISTED_BOOKMARK_VERSION)
                            // don't remove from the map until we've PERSISTED
                            toRemove.add(bmStr);
                        else
                            subscriptionMap.remove(bmStr);
                        if (subscriptionMapEntry > 0)
                        {
                            subscription.discard(subscriptionMapEntry);
                        }
                    }
                    break;

                case ENTRY_PERSISTED:
                    len = in.readUnsignedByte();
                    position++;
                    if(len == -1 || len > (_file.length() - position))
                    {
                        // truncated entry
                        _file.seek(entryStartPos);
                        in.close();
                        return;
                    }
                    in.readFully(bookmark, 0, len);
                    position += len;
                    bookmarkField.set(bookmark, 0, len);
                    bmStr = bookmarkField.getValue(_decoder);

                    if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION)
                    {
                        subscription.setLastPersisted(bookmarkField);
                    }
                    else if(subscriptionMap.containsKey(bmStr))
                    {
                        long discardIndex = subscriptionMap.get(bmStr);
                        subscription.persisted(discardIndex);
                        // empty out the subscriptionMap for everything in the toRemove.
                        Iterator<String> it = toRemove.iterator();
                        while (it.hasNext())
                        {
                            String bf = it.next();
                            Long subMapEntry = subscriptionMap.get(bf);
                            if(subMapEntry!=null && subMapEntry <= discardIndex)
                            {
                                it.remove();
                                subscriptionMap.remove(bf);
                            }
                        }
                    }
                    break;

                default:
                    throw new IOException("Corrupt file found.");
                }
                in.readUnsignedByte();
                position++;
                lastGoodPosition = position;
            }
            if (_file.getFilePointer() != lastGoodPosition)
                _file.seek(lastGoodPosition);
            in.close();
        }
        catch (IOException ex)
        {
            boolean onLastTruncatedLine = true;
            try {
                // Try to determine if we are on the last line of the file,
                // which may have been corrupted due to truncation from an
                // abrupt exit.
                int len = 0;
                byte[] buffer = new byte[255];
                while ((len = in.read(buffer, 0, buffer.length)) != -1) {
                    // Scan for a newline char from the current position.
                    for (int i = 0; i < len; i++) {
                        // If we found a newline, then we're not on the
                        // last line of the file that has been truncated.
                        if ((byte)'\n' == buffer[i]) {
                            onLastTruncatedLine = false;
                            break;
                        }
                    }
                }

                // Now close the stream.
                in.close();
            }
            catch (IOException ex2) {
                ex.addSuppressed(ex2);
                throw ex; // Something seems to be seriously wrong, so throw.
            }
            if (lastGoodPosition > 0 && onLastTruncatedLine)
            {
                // We only want to seek to the last good position if
                // we know we're on the last line of the file and it
                // has been truncated (due to an abrupt shutdown).
                try {
                    _file.seek(lastGoodPosition);
                }
                catch (IOException ex2) {
                    ex.addSuppressed(ex2);
                    throw ex;
                }
            }
            else
            {
                // The corruption seems to be somewhere earlier in the
                // file, so throw the exception.
                throw ex;
            }
        }
        finally
        {
            try {
                if (in != null) in.close();
            }
            catch (IOException ex) {
            }
            // Because this is only called either in constructor or in
            // setServerVersion already under the _subsLock, there is
            // no reason to lock _subsLock here.
            Iterator<Map.Entry<Field, Subscription>> it = _subs.entrySet().iterator();
            while (it.hasNext())
            {
                Map.Entry<Field, Subscription> pairs = it.next();
                boolean disposeSub = pairs.getValue().justRecovered();
                if (disposeSub)
                {
                    it.remove();
                    _pool.returnToPool(pairs.getValue());
                }
            }
            _recovering = false;
        }
    }

    public long log(Message message) throws AMPSException
    {
        BookmarkField bookmark = (BookmarkField)message.getBookmarkRaw();
        if (bookmark.equals(Subscription.EPOCH_FIELD)) return 0;
        if (bookmark.isTimestamp()
                && getServerVersion() >= Client.MIN_MULTI_BOOKMARK_VERSION)
        {
            return 0;
        }
        Subscription sub = (LoggedBookmarkStore.Subscription)message.getSubscription();
        Field subId = message.getSubIdRaw();
        if (subId == null || subId.isNull())
            subId = message.getSubIdsRaw();

        _lock.lock();
        try
        {
            if(_file == null)
            {
                throw new StoreException("Store not open.");
            }
        }
        finally
        {
            _lock.unlock();
        }

        long index = 0;
        try
        {
            if (sub == null)
            {
                sub = find(subId);
                message.setSubscription(sub);
            }
            index = sub.log(bookmark);
        }
        catch (IOException ioex)
        {
            throw new AMPSException("Error logging to bookmark store", ioex);
        }
        message.setBookmarkSeqNo(index);
        boolean callPersisted;
        _lock.lock();
        try
        {
            // log the arrival of this bookmark.
            write(subId, LoggedBookmarkStore.ENTRY_BOOKMARK, bookmark);
            callPersisted = _noPersistedAcks.contains(subId);
        }
        catch (IOException ioex)
        {
            throw new AMPSException("Error logging to bookmark store", ioex);
        }
        finally
        {
            _lock.unlock();
        }
        if (callPersisted)
            persisted(subId, bookmark);
        return index;
    }

    public void discard(Field subId, long bookmarkSeqNo) throws AMPSException
    {
        _lock.lock();
        try
        {
            if(_file == null)
            {
                throw new StoreException("Store not open.");
            }
        }
        finally
        {
            _lock.unlock();
        }
        try
        {
            find(subId).discard(bookmarkSeqNo);
        }
        catch (IOException ioex)
        {
            throw new AMPSException("Error discarding from bookmark store", ioex);
        }
    }

    public void discard(Message message) throws AMPSException
    {
        _lock.lock();
        try
        {
            if(_file == null)
            {
                throw new StoreException("Store not open.");
            }
        }
        finally
        {
            _lock.unlock();
        }
        long bookmark = message.getBookmarkSeqNo();
        Subscription sub = (LoggedBookmarkStore.Subscription)message.getSubscription();
        if (sub == null)
        {
            if (message.getBookmarkRaw().equals(Subscription.EPOCH_FIELD)) return;
            if (message.getBookmarkRaw().isTimestamp()
                    && getServerVersion() >= Client.MIN_MULTI_BOOKMARK_VERSION)
            {
                return;
            }
            Field subId = message.getSubIdRaw();
            if (subId == null || subId.isNull())
                subId = message.getSubIdsRaw();
            try
            {
                sub = find(subId);
            }
            catch (IOException ioex)
            {
                throw new AMPSException("Error discarding from bookmark store", ioex);
            }
            message.setSubscription(sub);
        }
        try
        {
            sub.discard(bookmark);
        }
        catch (IOException ioex)
        {
            throw new AMPSException("Error discarding from bookmark store", ioex);
        }
    }

    public Field getMostRecent(Field subId) throws AMPSException
    {
        _lock.lock();
        try
        {
            if(_file == null)
            {
                throw new StoreException("Store not open.");
            }
        }
        finally
        {
            _lock.unlock();
        }
        try
        {
            if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION)
                return find(subId).getMostRecentList();
            else
                return find(subId).getMostRecent();
        }
        catch (IOException ioex)
        {
            throw new AMPSException("Error getting most recent in bookmark store", ioex);
        }
    }

    public boolean isDiscarded(Message message) throws AMPSException
    {
        _lock.lock();
        try
        {
            if(_file == null)
            {
                throw new StoreException("Store not open.");
            }
        }
        finally
        {
            _lock.unlock();
        }
        BookmarkField bookmark = (BookmarkField)message.getBookmarkRaw();
        if (bookmark.equals(Subscription.EPOCH_FIELD)) return true;
        if (bookmark.isTimestamp()) return false;
        Subscription sub = (LoggedBookmarkStore.Subscription)message.getSubscription();
        if (sub == null)
        {
            Field subId = message.getSubIdRaw();
            if (subId == null || subId.isNull())
                subId = message.getSubIdsRaw();
            try
            {
                sub = find(subId);
            }
            catch (IOException ioex)
            {
                throw new AMPSException("Error checking is discarded in bookmark store", ioex);
            }
            message.setSubscription(sub);
        }
        try
        {
            return sub.isDiscarded(bookmark);
        }
        catch (IOException ioex)
        {
            throw new AMPSException("Error checking is discarded in bookmark store", ioex);
        }
    }

    public void persisted(Field subId, long bookmark) throws AMPSException
    {
        _lock.lock();
        try
        {
            if(_file == null)
            {
                throw new StoreException("Store not open.");
            }
        }
        finally
        {
            _lock.unlock();
        }
        try
        {
            if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION)
                find(subId).setLastPersisted(bookmark);
            else
                find(subId).persisted(bookmark);
        }
        catch (IOException ioex)
        {
            throw new AMPSException("Error logging persisted to bookmark store", ioex);
        }
    }

    public void persisted(Field subId, BookmarkField bookmark) throws AMPSException
    {
        _lock.lock();
        try
        {
            if(_file == null)
            {
                throw new StoreException("Store not open.");
            }
        }
        finally
        {
            _lock.unlock();
        }
        if (bookmark.equals(Subscription.EPOCH_FIELD)) return;
        try
        {
            if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION)
                find(subId).setLastPersisted(bookmark);
            else
                find(subId).persisted(bookmark);
        }
        catch (IOException ioex)
        {
            throw new AMPSException("Error logging persisted to bookmark store", ioex);
        }
    }

    public void noPersistedAcks(Field subId) throws AMPSException
    {
        try
        {
            Subscription sub = find(subId);
            _noPersistedAcks.add(subId.copy());
            sub.noPersistedAcks();
        }
        catch (IOException ioex)
        {
            throw new AMPSException("Error setting no persisted acks on bookmark store", ioex);
        }
    }

    public long getOldestBookmarkSeq(Field subId) throws AMPSException
    {
        _lock.lock();
        try
        {
            if(_file == null)
            {
                throw new StoreException("Store not open.");
            }
        }
        finally
        {
            _lock.unlock();
        }
        try
        {
            return find(subId).getOldestBookmarkSeq();
        }
        catch (IOException ioex)
        {
            throw new AMPSException("Error getting oldest bookmark seq from bookmark store", ioex);
        }
    }

    public void setResizeHandler(BookmarkStoreResizeHandler handler)
    {
        // The _resizeHandler is only touched under _subsLock
        _subsLock.lock();
        try {
            _resizeHandler = handler;
            Iterator it = _subs.entrySet().iterator();
            while (it.hasNext())
            {
                Map.Entry pairs = (Map.Entry)it.next();
                ((Subscription)pairs.getValue()).setResizeHandler(handler, this);
            }
        }
        finally { 
            _subsLock.unlock();
        }
    }

    private Subscription find(Field subId) throws IOException
    {
        _subsLock.lock();
        try {
            Subscription s = _subs.get(subId);
            if(s==null)
            {
                s=_pool.get();
                s.init(subId, this);
                s.setResizeHandler(_resizeHandler, this);
                if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION ||
                    _serverVersion < Client.MIN_PERSISTED_BOOKMARK_VERSION)
                {
                    s.noPersistedAcks();
                }
                _subs.put(subId.copy(), s);
            }
            return s;
        }
        finally {
            _subsLock.unlock();
        }
    }


    /*
    public void compactTo(String path) throws IOException
    {
        // Create the file
        // for each Subscription s in self:
            // write a ENTRY_BOOKMARK for s.getMostRecent()
            // write a ENTRY_DISCARD for s.getMostRecent()

            // for each entry e in s:
                // write a ENTRY_BOOKMARK for e
                // if !e._active: write a ENTRY_DISCARD for e
    }
    */


    /**
     *  Remove all entries in the bookmark store, completely
     *  clearing all record of messages received and discarded.
     */
    public void purge() throws AMPSException
    {
        _subsLock.lock();
        _lock.lock();
        try {
            // delete the file on disk.
            try
            {
                if(_file == null)
                {
                    throw new StoreException("Store not open.");
                }
                _file.setLength(0);
            }
            catch(IOException ioex)
            {
                throw new StoreException("Error truncating file", ioex);
            }

            for (Subscription sub: _subs.values())
            {
                sub.reset();
                _pool.returnToPool(sub);
            }

            _subs.clear();
            _noPersistedAcks.clear();
        }
        finally { 
            _lock.unlock();
            _subsLock.unlock();
        }
    }

    public void purge(Field subId_) throws AMPSException
    {
        _subsLock.lock();
        _lock.lock();
        String name = "";
        try {
            if(_file == null)
            {
                throw new StoreException("Store not open.");
            }
            Subscription sub = _subs.remove(subId_);
            if (sub == null) return;
            sub.reset();
            _pool.returnToPool(sub);
            _noPersistedAcks.remove(subId_);
            name = _fileName + ".tmp";
            prune(name);
        }
        catch (IOException e) {
            throw new StoreException("Underlying IOException while pruning. "
                    + "temp bookmark log file = " + name, e);
        }
        finally {
            _lock.unlock();
            _subsLock.unlock();
        }
    }
    
    /**
    * Closes the bookmark store. If it is already closed, a StoreException is thrown. 
    * @throws StoreException if there is an error closing the file backing in the store, 
    * or the store already closed.
    */
    public void close() throws AMPSException
    {
        _lock.lock();
        try {
            try
            {
                if(_file != null)
                {
                    _file.close();
                }
                else
                {
                    throw new StoreException("Store not open.");
                }
            }
            catch(IOException ioex)
            {
                throw new StoreException("Error closing file", ioex);
            }
            finally
            {
                _file = null;
            }
        }
        finally { 
            _lock.unlock();
        }
    }

    /**
     * Used to change the version of the AMPS server that this bookmark store's
     * client has connected to.
     * 
     * @param version An AMPS server version integer of the form 03080000 for
     *        version 3.8.0.0.
     */
    public void setServerVersion(int version)
    {
        _subsLock.lock();
        _lock.lock();
        try {
            // Check for no change in version
            if (_serverVersion == version) return;

            // Note whether the old and new versions are in the
            // "persisted version window" (i.e. v3.8 >= version < v4.0).
            boolean oldVerInPersistWindow = false;
            boolean newVerInPersistWindow = false;
            if (_serverVersion >= Client.MIN_PERSISTED_BOOKMARK_VERSION
                    && _serverVersion < Client.MIN_MULTI_BOOKMARK_VERSION)
                oldVerInPersistWindow = true;
            if (version >= Client.MIN_PERSISTED_BOOKMARK_VERSION
                    && version < Client.MIN_MULTI_BOOKMARK_VERSION)
                newVerInPersistWindow = true;
            
            // Now change version
            _serverVersion = version;
            
            // Check for no functional change in version
            if (oldVerInPersistWindow == newVerInPersistWindow) return;
            
            if(_file == null) return;
            if (!newVerInPersistWindow)
            {
                Iterator it = _subs.entrySet().iterator();
                while (it.hasNext())
                {
                    Map.Entry pairs = (Map.Entry)it.next();
                    ((Subscription)pairs.getValue()).noPersistedAcks();
                }
            }
            else
            {
                // If version was changed to this, we need to re-recover
                for (Subscription sub: _subs.values())
                {
                    sub.reset();
                    _pool.returnToPool(sub);
                }
                _subs.clear();
                // This didn't throw the first time, so should be fine
                try {
                    _file.close();
                    _file = new RandomAccessFile(_fileName, "rw");
                    recover();
                }
                catch(IOException e) {
                    throw new RuntimeException("The bookmark store file "
                            + _fileName + " has encountered an error.", e);
                }
            }
        }
        finally { 
            _lock.unlock();
            _subsLock.unlock();
        }
    }
    
    /**
    * Called by Client when connected to an AMPS server in order to retrieve the version 
    * number of the server. Returns retrieved version number. 
    * @return the server version, represented as an integer
    */
    public int getServerVersion()
    {
        _lock.lock();
        try {
            return _serverVersion;
        }
        finally { 
            _lock.unlock();
        }
    }


    /**
     * Locks and returns the current list of Subscriptions.
     * Used internally by prune() to gather and lock the list of subscriptions
     * that we need to prune. We hold self's lock here to get the subscription 
     * list and then release it to lock up all of the subscriptions.
     */
    private List<Map.Entry<Field,Subscription>> _lockSubs()
    {
        // The _subsLock MUST already be held
        // No need to hold self's _lock here.
        List<Map.Entry<Field,Subscription>> subs
            = new Vector<Map.Entry<Field,Subscription>>(_subs.size());
        for(Map.Entry<Field,Subscription> entry : _subs.entrySet())
        {
            subs.add(entry);
        }

        // Lock the subscriptions without self's lock held.
        for(Map.Entry<Field,Subscription> entry : subs)
        {
            entry.getValue().lock();
        }
        return subs;
    }

    /**
     * Unlocks the provided list of Subscriptions.
     * Used internally by prune() to gather and lock the list of subscriptions
     * that we need to prune. We hold self's lock here to get the subscription 
     * list and then release it to lock up all of the subscriptions.
     */
    private void _unlockSubs(List<Map.Entry<Field,Subscription>> subs_)
    {
        // The _subsLock MUST already be held
        // No need to hold self's _lock here.
        // We are just unlocking the individual Subscriptions that were
        // held for the duration of prune().
        for(Map.Entry<Field,Subscription> entry : subs_)
        {
            entry.getValue().unlock();
        }
    }
}
