///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2019 60East Technologies Inc., All Rights Reserved.
//
// This computer software is owned by 60East Technologies Inc. and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// Inc. pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies Inc.; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies Inc.
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government's use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies Inc..
//
////////////////////////////////////////////////////////////////////////////

package com.crankuptheamps.client;

import java.io.IOException;
import java.lang.Long;
import java.math.BigInteger;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import com.crankuptheamps.client.fields.Field;
import com.crankuptheamps.client.fields.BookmarkField;
import com.crankuptheamps.client.LoggedBookmarkStore.Subscription;
import com.crankuptheamps.client.exception.*;


/** Implements a bookmark store useful for handling
 * server failover scenarios, but without a backing store to recover from subscriber failure.
 */
public class MemoryBookmarkStore implements BookmarkStore
{
    // We use an in-memory array to remember the order in which bookmarks
    // arrive on a given subscription.  When the user has discard()ed all of the
    // bookmarks up to a certain bookmark B, we log B, and return it the next time
    // someone calls getMostRecent().
    static class Subscription implements com.crankuptheamps.client.Subscription
    {
        // This subscription's ID.
        Field _sub;

        // The last persisted bookmark
        private BookmarkField _lastPersisted;

        static final Field EPOCH_FIELD = new Field(Client.Bookmarks.EPOCH);

        BookmarkRingBuffer _ring = new BookmarkRingBuffer();

        // A set of all of the undiscarded entries recovered that
        // had been received after the one returned by getMostRecent().
        // When these come in again, we won't log them a 2nd time.
        HashMap<Field, Long> _recovered = new HashMap<Field, Long>();

        // The per-subscription memory of what we've seen from publishers
        HashMap<Long, Long> _publishers = new HashMap<Long, Long>();

        // A map of active bookmark indices to their new location after a reconnect
        HashMap<Long, Long> _relocatedActiveBookmarks = new HashMap<Long, Long>();

        // The Subscription lock
        private final Lock       lock           = new ReentrantLock();

        MemoryBookmarkStore _parent;

        public Subscription(Field subscriptionId)
        {
            _sub = subscriptionId.copy();
            _ring.setSubId(_sub);
            _lastPersisted = new BookmarkField();
            _lastPersisted.copyFrom(EPOCH_FIELD);
        }
        public Subscription()
        {
            _lastPersisted = new BookmarkField();
            _lastPersisted.copyFrom(EPOCH_FIELD);
        }
        public void setSubscription(Field subId)
        {
            _sub = subId.copy();
            _ring.setSubId(_sub);
        }
        public void setParent(MemoryBookmarkStore parent)
        {
            _parent = parent;
        }

        public long log(BookmarkField bookmark)
        {
            lock.lock();
            try {
                // Check to see if this is a recovered bookmark.
                // If so, just return that one -- don't re-log.
                final Long recoveredValue = _recovered.remove(bookmark);
                if (recoveredValue != null) {
                    long seq = _ring.relog(recoveredValue, bookmark);
                    while (seq == 0) {
                        lock.unlock();
                        try {
                            _ring.checkResize();
                        }
                        finally {
                            lock.lock();
                        }
                        seq = _ring.relog(recoveredValue, bookmark);
                    }
                    _relocatedActiveBookmarks.put(recoveredValue, seq);
                    return seq;
                }
                long seq = _ring.log(bookmark);
                while (seq == 0) {
                    lock.unlock();
                    try {
                        _ring.checkResize();
                    }
                    finally {
                        lock.lock();
                    }
                    seq = _ring.log(bookmark);
                }
                return seq;
            }
            finally {
                lock.unlock();
            }
        }

        public void discard(long index)
        {
            lock.lock();
            try {
                long idx = index;
                Long newIndex = _relocatedActiveBookmarks.remove(index);
                // In case something's been recovered more than once, get full indirection
                while (newIndex != null)
                {
                    _ring.discard(idx);
                    idx = newIndex;
                    newIndex = _relocatedActiveBookmarks.remove(newIndex);
                }
                _ring.discard(idx);
            }
            finally {
                lock.unlock();
            }
        }

        // Check to see if this message is older than the most recent one seen,
        // and if it is, check if it discarded.
        public boolean isDiscarded(BookmarkField bookmark)
        {
            lock.lock();
            try {
                Long recoveredIndex = _recovered.get(bookmark);
                if (recoveredIndex != null)
                {
                    BookmarkRingBuffer.Entry entry = _ring.getByIndex(recoveredIndex);
                    boolean active = (entry == null) ? false : entry.isActive();
                    if (active)
                    {
                        return false;
                    }
                    if ((entry == null ||
                        (!entry.isActive() && entry.isPersisted())) &&
                        _ring.getStartIndex() == BookmarkRingBuffer.UNSET_INDEX)
                    {
                        // This will update most recent if necessary
                        _ring.discard(recoveredIndex);
                    }
                    else
                    {
                        long seq = _ring.relog(recoveredIndex, bookmark);
                        while (seq == 0) {
                            lock.unlock();
                            try {
                                _ring.checkResize();
                            }
                            finally {
                                lock.lock();
                            }
                            seq = _ring.relog(recoveredIndex, bookmark);
                        }
                    }
                    _recovered.remove(bookmark);
                    return true;
                }
                long publisher = bookmark.getPublisherId();
                long sequence = bookmark.getSequenceNumber();
                if(!_publishers.containsKey(publisher) ||
                    unsignedLongLess(_publishers.get(publisher), sequence))
                {
                    _publishers.put(publisher, sequence);
                    return false;
                }
                BookmarkRingBuffer.Entry entry =
                        _ring.find(bookmark);
                if(entry != null)
                {
                    return !entry.isActive();
                }

                return true; // message is totally discarded
            }
            finally {
                lock.unlock();
            }
        }

        public Field getMostRecent()
        {
            lock.lock();
            try {
                // _recent is the most recent bookmark.
                // when this is called, we'll take a moment to update the
                // list of things recovered, so we don't accidentally log
                // anything we ought not to.
                updateRecovery();
                return _ring.getLastDiscarded();
            }
            finally {
                lock.unlock();
            }
        }

        // publisherId_ and sequence are signed, but the actual id's are
        // unsigned, so we have to do some work to create a valid string.
        private static String convertUnsignedLongToString(long publisherId_)
        {
            final BigInteger offset = BigInteger.valueOf(Long.MAX_VALUE)
                    .shiftLeft(1).add(BigInteger.valueOf(2));
            if(publisherId_<0)
            {
                return offset.add(BigInteger.valueOf(publisherId_)).toString();
            }
            else
            {
                return Long.toString(publisherId_);
            }
        }

        /** publisherId_ and sequence are signed, but the actual id's are
         * unsigned, so we have to do some work to compare them.
         */
        private static boolean unsignedLongLess(long seqLeft_, long seqRight_)
        {
            final BigInteger offset = BigInteger.valueOf(Long.MAX_VALUE)
                    .shiftLeft(1).add(BigInteger.valueOf(2));
            if(seqLeft_<0 || seqRight_<0)
            {
                BigInteger left = offset.add(BigInteger.valueOf(seqLeft_));
                BigInteger right = offset.add(BigInteger.valueOf(seqRight_));
                return left.compareTo(right) < 0;
            }
            else
            {
                return seqLeft_ < seqRight_;
            }
        }

        public Field getMostRecentList()
        {
            final CharsetEncoder encoder = StandardCharsets.UTF_8.newEncoder();
            final CharsetDecoder decoder = StandardCharsets.UTF_8.newDecoder();
            lock.lock();
            try {
                // when this is called, we'll take a moment to update the list
                // of things recovered, so we don't accidentally log anything
                // we ought not to.
                updateRecovery();
                BookmarkField lastDiscarded = (BookmarkField)_ring.getLastDiscarded();
                boolean useLastDiscarded = (lastDiscarded != null &&
                                            !lastDiscarded.isNull());
                long lastDiscardedPub = 0;
                long lastDiscardedSeq = 0;
                boolean useLastPersisted = (_lastPersisted != null &&
                                            !_lastPersisted.isNull());
                long lastPersistedPub = 0;
                long lastPersistedSeq = 0;
                if (useLastPersisted)
                {
                    lastPersistedPub = _lastPersisted.getPublisherId();
                    lastPersistedSeq = _lastPersisted.getSequenceNumber();
                }
                if (useLastDiscarded)
                {
                    if (_ring.isEmpty() && useLastPersisted)
                    {
                        useLastDiscarded = false;
                    }
                    else
                    {
                        lastDiscardedPub = lastDiscarded.getPublisherId();
                        lastDiscardedSeq = lastDiscarded.getSequenceNumber();
                        // Only use one if they are the same publisher
                        if (useLastPersisted &&
                            (lastPersistedPub == lastDiscardedPub))
                        {
                            useLastPersisted = (lastPersistedSeq < lastDiscardedSeq);
                            useLastDiscarded = !useLastPersisted;
                        }
                    }
                }
                BookmarkField recentList = new BookmarkField();
                StringBuilder recentStr = new StringBuilder();
                if (useLastDiscarded)
                {
                    recentStr.append((lastDiscarded).getValue(decoder));
                }
                // If we don't have a last persisted or a last discarded OR we
                // are expecting persisted acks but haven't received one yet,
                // then we should try to build a list of bookmarks based on
                // publishers we have seen so far, if any, or return EPOCH.
                if ((!useLastPersisted && !useLastDiscarded) ||
                    (_lastPersisted != null &&
                     _lastPersisted.equals(EPOCH_FIELD)))
                {
                    if (_publishers.isEmpty())
                    {
                        // Set last persisted to EPOCH and return it
                        if (_lastPersisted == null) {
                            _lastPersisted = new BookmarkField();
                            _lastPersisted.copyFrom(EPOCH_FIELD);
                        }
                        return _lastPersisted;
                    }
                    if (recentStr.length() > 0) recentStr.setLength(0);
                    Iterator it = _publishers.entrySet().iterator();
                    while (it.hasNext())
                    {
                        if (recentStr.length() > 0) recentStr.append(',');
                        Map.Entry pairs = (Map.Entry)it.next();
                        long pubId = (Long)pairs.getKey();
                        long seq = (Long)pairs.getValue();
                        recentStr.append(convertUnsignedLongToString(pubId))
                                 .append(BookmarkField.SEPARATOR_CHAR)
                                 .append(convertUnsignedLongToString(seq))
                                 .append(BookmarkField.SEPARATOR_CHAR);
                    }
                    recentList.setValue(recentStr.toString(), encoder);
                    return recentList;
                }
                if (useLastPersisted)
                {
                    if (recentStr.length() > 0) recentStr.append(',');
                    recentStr.append(_lastPersisted.getValue(decoder));
                }
                recentList.setValue(recentStr.toString(), encoder);
                return recentList;
            }
            finally {
                lock.unlock();
            }
        }

        private void updateRecovery()
        {
            long end = _ring.getEndIndex();
            for (long index = _ring.setRecovery(); index < end; ++index)
            {
                BookmarkRingBuffer.Entry entry = _ring.getByIndex(index);
                if(entry != null && (entry.isActive() || !entry.isPersisted()))
                {
                    _recovered.put(entry.getBookmark(), index);
                }
            }
        }

        public void persisted(long bookmark)
        {
            lock.lock();
            try {
            _ring.persisted(bookmark);
            }
            finally {
                lock.unlock();
            }
        }

        public void persisted(BookmarkField bookmark)
        {
            lock.lock();
            try {
                _ring.persisted(bookmark);
            }
            finally {
                lock.unlock();
            }
        }

        public void setLastPersisted(long bookmark)
        {
            lock.lock();
            try {
                BookmarkRingBuffer.Entry entry = _ring.getByIndex(bookmark);
                if (entry == null)
                    return;
                if (_lastPersisted != null) _lastPersisted.reset();
                _lastPersisted = entry.getBookmark().copy();
            }
            finally {
                lock.unlock();
            }
        }

        public void setLastPersisted(BookmarkField bookmark)
        {
            lock.lock();
            try {
                if (bookmark == null || bookmark.isNull()) return;

                if (bookmark.equals(_lastPersisted)) return;

                if (bookmark.isTimestamp())
                {
                    if (_lastPersisted != null) _lastPersisted.reset();
                    _lastPersisted = bookmark.copy();
                    return;
                }

                if (_lastPersisted != null && !_ring.persistedAcks() && 
                    bookmark.getPublisherId() == _lastPersisted.getPublisherId() &&
                    bookmark.getSequenceNumber() <= _lastPersisted.getSequenceNumber())
                {
                    return;
                }
                if (_lastPersisted != null) _lastPersisted.reset();
                _lastPersisted = bookmark.copy();
            }
            finally {
                lock.unlock();
            }
        }

        public void noPersistedAcks()
        {
            lock.lock();
            try {
                if (_parent.getServerVersion() < Client.MIN_PERSISTED_BOOKMARK_VERSION ||
                    _parent._noPersistedAcks.contains(_sub))
                    _lastPersisted = null;
                _ring.noPersistedAcks();
            }
            finally {
                lock.unlock();
            }
        }

        public void setPersistedAcks()
        {
            lock.lock();
            try {
                _lastPersisted = new BookmarkField();
                _lastPersisted.copyFrom(EPOCH_FIELD);
                _ring.setPersistedAcks();
            }
            finally {
                lock.unlock();
            }
        }

        public long getOldestBookmarkSeq()
        {
            lock.lock();
            try {
                return _ring.getStartIndex();
            }
            finally {
                lock.unlock();
            }
        }

        public void setResizeHandler(BookmarkStoreResizeHandler handler, BookmarkStore store)
        {
            _ring.setResizeHandler(handler, store);
        }
    }

    HashMap<Field, Subscription> _subs = new HashMap<Field, Subscription>();
    HashSet<Field> _noPersistedAcks = new HashSet<Field>();
    BookmarkStoreResizeHandler _resizeHandler = null;
    private int _serverVersion = Client.MIN_MULTI_BOOKMARK_VERSION;
    // The Store lock
    private final Lock       lock           = new ReentrantLock();

    // Subscriptions take a while to initialize, so we avoid doing this in the message handler.
    // number of subscriptions we pre-initialize.
    Pool<Subscription> _pool;
    public MemoryBookmarkStore()
    {
        _pool = new Pool<Subscription>(Subscription.class, 1);
    }

    /**
     * Initialize self with a target number of subscriptions to store
     * @param targetNumberOfSubscriptions the number of subscriptions this store will need to track
     */
    public MemoryBookmarkStore(int targetNumberOfSubscriptions)
    {
        _pool = new Pool<Subscription>(Subscription.class, targetNumberOfSubscriptions);
    }
    public long log(Message message) throws AMPSException
    {
        BookmarkField bookmark = (BookmarkField)message.getBookmarkRaw();
        if (bookmark.equals(Subscription.EPOCH_FIELD)) return 0;
        if (bookmark.isTimestamp()
                && getServerVersion() >= Client.MIN_MULTI_BOOKMARK_VERSION)
        {
            return 0;
        }
        Subscription sub = (MemoryBookmarkStore.Subscription)message.getSubscription();
        if (sub == null)
        {
            Field subId = message.getSubIdRaw();
            if (subId == null || subId.isNull())
                subId = message.getSubIdsRaw();
            sub = find(subId);
            message.setSubscription(sub);
        }
        long seqNo = sub.log(bookmark);
        message.setBookmarkSeqNo(seqNo);
        return seqNo;
    }

    public void discard(Field subId, long bookmarkSeqNo) throws AMPSException
    {
        find(subId).discard(bookmarkSeqNo);
    }

    public void discard(Message message) throws AMPSException
    {
        long bookmark = message.getBookmarkSeqNo();
        Subscription sub = (MemoryBookmarkStore.Subscription)message.getSubscription();
        if (sub == null)
        {
            if (message.getBookmarkRaw().equals(Subscription.EPOCH_FIELD)) return;
            if (message.getBookmarkRaw().isTimestamp()
                    && getServerVersion() >= Client.MIN_MULTI_BOOKMARK_VERSION)
            {
                return;
            }
            Field subId = message.getSubIdRaw();
            if (subId == null || subId.isNull())
                subId = message.getSubIdsRaw();
            sub = find(subId);
            message.setSubscription(sub);
        }
        sub.discard(bookmark);
    }

    public Field getMostRecent(Field subId) throws AMPSException
    {
        if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION)
            return find(subId).getMostRecentList();
        else
            return find(subId).getMostRecent();
    }

    public boolean isDiscarded(Message message) throws AMPSException
    {
        BookmarkField bookmark = (BookmarkField)message.getBookmarkRaw();
        if (bookmark.equals(Subscription.EPOCH_FIELD)) return true;
        if (bookmark.isTimestamp()) return false;
        Subscription sub = (MemoryBookmarkStore.Subscription)message.getSubscription();
        if (sub == null)
        {
            Field subId = message.getSubIdRaw();
            if (subId == null || subId.isNull())
                subId = message.getSubIdsRaw();
            sub = find(subId);
            message.setSubscription(sub);
        }
        return sub.isDiscarded(bookmark);
    }

    public void persisted(Field subId, long bookmark) throws AMPSException
    {
        if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION)
            find(subId).setLastPersisted(bookmark);
        else
            find(subId).persisted(bookmark);
    }

    public void persisted(Field subId, BookmarkField bookmark) throws AMPSException
    {
        if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION)
            find(subId).setLastPersisted(bookmark);
        else
            find(subId).persisted(bookmark);
    }

    public void noPersistedAcks(Field subId) throws AMPSException
    {
        Subscription sub = find(subId);
        _noPersistedAcks.add(subId.copy());
        sub.noPersistedAcks();
    }

    private Subscription find(Field subId)
    {
        lock.lock();
        try {
            Subscription s = _subs.get(subId);
            if(s==null)
            {
                s = _pool.get();
                s.setSubscription(subId);
                s.setParent(this);
                s.setResizeHandler(_resizeHandler, this);
                if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION ||
                    _serverVersion < Client.MIN_PERSISTED_BOOKMARK_VERSION)
                    s.noPersistedAcks();
                _subs.put(subId.copy(), s);
            }
            return s;
        }
        finally {
            lock.unlock();
        }
    }

    public void purge() throws AMPSException
    {
        lock.lock();
        try {
            _subs = new HashMap<Field, Subscription>();
            _noPersistedAcks = new HashSet<Field>();
        }
        finally {
            lock.unlock();
        }
    }

    public void purge(Field subId_) throws AMPSException
    {
        lock.lock();
        try {
            if (_subs.remove(subId_) == null) return;
            _noPersistedAcks.remove(subId_);
        }
        finally {
            lock.unlock();
        }
    }

    public long getOldestBookmarkSeq(Field subId) throws AMPSException
    {
        lock.lock();
        try {
            long retVal = 0;
            retVal = find(subId).getOldestBookmarkSeq();
            return retVal;
        }
        finally {
            lock.unlock();
        }
    }

    public void setResizeHandler(BookmarkStoreResizeHandler handler)
    {
        _resizeHandler = handler;
        Iterator it = _subs.entrySet().iterator();
        while (it.hasNext())
        {
            Map.Entry pairs = (Map.Entry)it.next();
            ((Subscription)pairs.getValue()).setResizeHandler(handler, this);
        }
    }

    public void setServerVersion(int version)
    {
        if (_serverVersion == version) return;
        _serverVersion = version;
        if (_serverVersion >= Client.MIN_MULTI_BOOKMARK_VERSION ||
            _serverVersion < Client.MIN_PERSISTED_BOOKMARK_VERSION)
        {
            Iterator it = _subs.entrySet().iterator();
            while (it.hasNext())
            {
                Map.Entry pairs = (Map.Entry)it.next();
                ((Subscription)pairs.getValue()).noPersistedAcks();
            }
        }
        else
        {
            Iterator it = _subs.entrySet().iterator();
            while (it.hasNext())
            {
                Map.Entry pairs = (Map.Entry)it.next();
                ((Subscription)pairs.getValue()).setPersistedAcks();
            }
        }
    }

    /**
    * Called by Client when connected to an AMPS server in order to retrieve the version 
    * number of the server. Returns retrieved version number. 
    * @return the server version, represented as an integer
    */

    public int getServerVersion()
    {
        return _serverVersion;
    }
    
    
    public void close()
    {
        // no-op
    }
}
